/************************************************************************

misc.c - miscellaneous helper functions

This file is a part of sscsss -- Simple Single Cycle Sample Set Synthesizer.

Copyright (C) 2002 David Baker

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "sscsss.h"

/************************************************************************

errex - log error message and exit

************************************************************************/

void errex(M_P m, int pnum, char *msg)
{
    if ( m->log == NULL ) {
	if ( pnum )
	  printf("Error (line %d):  %s\n", m->linecnt, msg);
	else if ( m->w_back != NULL )
	  printf("Error (wave %s):  %s\n", m->w_back->name, msg);
	else
	  printf("Error:  %s\n", msg);
    }
    else {
	if ( pnum )
	  fprintf(m->log, "Error (line %d):  %s\n", m->linecnt, msg);
	else if ( m->w_back != NULL )
	  fprintf(m->log, "Error (wave %s):  %s\n", m->w_back->name, msg);
	else
	  fprintf(m->log, "Error:  %s\n", msg);
	fclose(m->log);
    }
    exit(1);
}

/************************************************************************

free_lists - free memory used for lists, recursively

************************************************************************/

/* forward references */
void free_ws_list(WS_P, int*);
void free_o_list(O_P, int*);
void free_fi_list(FI_P, int*);
void free_gp_list(GP_P, int*);
void free_sc_list(SC_P, int*);
void free_fo_list(FO_P, int*);
void free_wp_list(WP_P, int*);
void free_note_list(N_P, int*);
void free_wpc_list(WPC_P, int*);

void free_w_list(W_P, int*);
void free_s_list(S_P, int*);
void free_h_list(H_P, int*);

void free_lists(M_P m)	/* free all lists starting at m */
{
    /* free wave sets */
    if ( m->ws_list != NULL ) {
	free_ws_list(m->ws_list, &(m->d_free));
	free(m->ws_list);
	(m->d_free)++;
	m->ws_list = NULL;
    }

    /* free waves */
    if ( m->w_list != NULL ) {
	free_w_list(m->w_list, &(m->d_free));
	free(m->w_list);
	(m->d_free)++;
	m->w_list = NULL;
    }
}

/* wave set list */
void free_ws_list(WS_P ws, int *freec) /* free wave set list */
{
    /* free following wave sets */
    if ( ws->next != NULL ) {
	free_ws_list(ws->next, freec);
	free(ws->next);
	(*freec)++;
    }

    /* free offset list */
    if ( ws->o_list != NULL ) {
	free_o_list(ws->o_list, freec);
	free(ws->o_list);
	(*freec)++;
    }

    /* free filter list */
    if ( ws->fi_list != NULL ) {
	free_fi_list(ws->fi_list, freec);
	free(ws->fi_list);
	(*freec)++;
    }

    /* free formant list */
    if ( ws->fo_list != NULL ) {
	free_fo_list(ws->fo_list, freec);
	free(ws->fo_list);
	(*freec)++;
    }

    /* free split list */
    if ( ws->wp_list != NULL ) {
	free_wp_list(ws->wp_list, freec);
	free(ws->wp_list);
	(*freec)++;
    }
}

void free_o_list(O_P of, int *freec) /* free offset lists */
{
    /* free following wave splits */
    if ( of->next != NULL ) {
	free_o_list(of->next, freec);
	free(of->next);
	(*freec)++;
    }

    /* free level scaling list */
    if ( of->level_s_list != NULL ) {
	free_sc_list(of->level_s_list, freec);
	free(of->level_s_list);
	(*freec)++;
    }

    /* free phase scaling list */
    if ( of->phase_s_list != NULL ) {
	free_sc_list(of->phase_s_list, freec);
	free(of->phase_s_list);
	(*freec)++;
    }
}

void free_fi_list(FI_P f, int *freec) /* free filter lists */
{
    /* free following filters */
    if ( f->next != NULL ) {
	free_fi_list(f->next, freec);
	free(f->next);
	(*freec)++;
    }

    /* free p1 scaling list */
    if ( f->p1_s_list != NULL ) {
	free_sc_list(f->p1_s_list, freec);
	free(f->p1_s_list);
	(*freec)++;
    }

    /* free p2 scaling list */
    if ( f->p2_s_list != NULL ) {
	free_sc_list(f->p2_s_list, freec);
	free(f->p2_s_list);
	(*freec)++;
    }

    /* free p3 scaling list */
    if ( f->p3_s_list != NULL ) {
	free_sc_list(f->p3_s_list, freec);
	free(f->p3_s_list);
	(*freec)++;
    }

    /* free p4 scaling list */
    if ( f->p4_s_list != NULL ) {
	free_sc_list(f->p4_s_list, freec);
	free(f->p4_s_list);
	(*freec)++;
    }

    /* free graph point list */
    if ( f->gp_list != NULL ) {
	free_gp_list(f->gp_list, freec);
	free(f->gp_list);
	(*freec)++;
    }
}

void free_sc_list(SC_P sc, int *freec) /* free scale lists */
{
    /* free following scale points */
    if ( sc->next != NULL ) {
	free_sc_list(sc->next, freec);
	free(sc->next);
	(*freec)++;
    }
}

void free_gp_list(GP_P gp, int *freec) /* free graph point list */
{
    /* free following graph points */
    if ( gp->next != NULL ) {
	free_gp_list(gp->next, freec);
	free(gp->next);
	(*freec)++;
    }
}

void free_fo_list(FO_P r, int *freec) /* free formant lists */
{
    /* free following formants */
    if ( r->next != NULL ) {
	free_fo_list(r->next, freec);
	free(r->next);
	(*freec)++;
    }

    /* free graph point list */
    if ( r->gp_list != NULL ) {
	free_gp_list(r->gp_list, freec);
	free(r->gp_list);
	(*freec)++;
    }
}

void free_wp_list(WP_P wp, int *freec) /* free wave split lists */
{
    /* free following wave splits */
    if ( wp->next != NULL ) {
	free_wp_list(wp->next, freec);
	free(wp->next);
	(*freec)++;
    }

    /* free wave split note list */
    if ( wp->note_list != NULL ) {
	free_note_list(wp->note_list, freec);
	free(wp->note_list);
	(*freec)++;
    }

    /* free wave split condition list */
    if ( wp->wpc_list != NULL ) {
	free_wpc_list(wp->wpc_list, freec);
	free(wp->wpc_list);
	(*freec)++;
    }
}

void free_note_list(N_P n, int *freec) /* free wave split note lists */
{
    /* free following wave split notes */
    if ( n->next != NULL ) {
	free_note_list(n->next, freec);
	free(n->next);
	(*freec)++;
    }
}

void free_wpc_list(WPC_P wpc, int *freec) /* free wave split condition lists */
{
    /* free following wave split conditions */
    if ( wpc->next != NULL ) {
	free_wpc_list(wpc->next, freec);
	free(wpc->next);
	(*freec)++;
    }
}

/* wave list */
void free_w_list(W_P w, int *freec)	/* free wave list */
{
    /* free following waves */
    if ( w->next != NULL ) {
	free_w_list(w->next, freec);
	free(w->next);
	(*freec)++;
    }

    /* free sample list */
    if ( w->s_list != NULL ) {
	free_s_list(w->s_list, freec);
	free(w->s_list);
	(*freec)++;
    }
}

void free_s_list(S_P s, int *freec)	/* free sample lists */
{
    /* free following samples */
    if ( s->next != NULL ) {
	free_s_list(s->next, freec);
	free(s->next);
	(*freec)++;
    }

    /* free harmonic list */
    if ( s->h_list != NULL ) {
	free_h_list(s->h_list, freec);
	free(s->h_list);
	(*freec)++;
    }

    /* free sample data */
    free(s->data);
    (*freec)++;
}

void free_h_list(H_P h, int *freec)	/* free harmonic lists */
{
    /* free following harmonics */
    if ( h->next != NULL ) {
	free_h_list(h->next, freec);
	free(h->next);
	(*freec)++;
    }
}

/************************************************************************

dump_m - debug dump of some stuff from the m(isc) structure

************************************************************************/

void dump_m(M_P m)
{
    fprintf(m->log, "\n\ndump_m\n\n");
    fprintf(m->log, "linecnt    = %d\n", m->linecnt);
    fprintf(m->log, "f_waves    = %d\n", m->f_waves);
    fprintf(m->log, "f_samples  = %d\n", m->f_samples);
    fprintf(m->log, "f_size     = %d\n", m->f_size);
    fprintf(m->log, "v_cat      = %s\n", m->v_cat);
    fprintf(m->log, "v_name     = %s\n", m->v_name);
    fprintf(m->log, "v_bend     = %d\n", m->v_bend);
    fprintf(m->log, "v_low      = %d\n", m->v_low);
    fprintf(m->log, "v_high     = %d\n", m->v_high);
    fprintf(m->log, "v_norm     = %d\n", m->v_norm);
    fprintf(m->log, "o_lowfreq  = %g\n", m->o_lowfreq);
    fprintf(m->log, "o_debug    = %d\n", m->o_debug);
    fprintf(m->log, "o_logws    = %d\n", m->o_logws);
    fprintf(m->log, "o_mkplot   = %d\n", m->o_mkplot);
    fprintf(m->log, "o_plotfreq = %d\n", m->o_plotfreq);
    fprintf(m->log, "o_plotpause= %d\n", m->o_plotpause);
    fprintf(m->log, "d_malloc   = %d\n", m->d_malloc);
    fprintf(m->log, "d_free     = %d\n", m->d_free);
    fprintf(m->log, "o_mindba   = %g\n", m->o_mindba);
}

/***********************************************************************

wave_set_list_dump

***********************************************************************/

/* forward references */
void offset_list_dump(M_P, O_P);
void filter_list_dump(M_P, FI_P);
void scale_list_dump(M_P, SC_P, char *);
void gp_list_dump(M_P, GP_P);
void formant_list_dump(M_P, FO_P);
void wave_split_list_dump(M_P, WP_P);
void wave_split_note_list_dump(M_P, N_P);
void wave_split_condition_list_dump(M_P, WPC_P);

void wave_set_list_dump(M_P m)
{
    WS_P ptr;

    fprintf(m->log, "\n\nwave_set_list_dump\n\n");
    for ( ptr = m->ws_list; ptr != NULL; ptr = ptr->next ) {
	offset_list_dump(m, ptr->o_list);
	filter_list_dump(m, ptr->fi_list);
	formant_list_dump(m, ptr->fo_list);
	wave_split_list_dump(m, ptr->wp_list);
    }
}

void offset_list_dump(M_P m, O_P of)
{
    for ( ; of != NULL; of = of->next ) {
	fprintf(m->log,
	  "  Offset:  harmonic %-4.4d level %-4.8g phase %-4.8g\n",
	  of->no, of->level, of->phase);
	scale_list_dump(m, of->level_s_list, "level");
	scale_list_dump(m, of->phase_s_list, "phase");
    }
}

void filter_list_dump(M_P m, FI_P f)
{
    for ( ; f != NULL; f = f->next ) {
	fprintf(m->log,
	  "  Filter:  type %1.1d p1 %-4.8g p2 %-4.8g p3 %-4.8g p4 %-4.8g\n",
	  f->type, f->p1, f->p2, f->p3, f->p4);
	scale_list_dump(m, f->p1_s_list, "p1");
	scale_list_dump(m, f->p2_s_list, "p2");
	scale_list_dump(m, f->p3_s_list, "p3");
	scale_list_dump(m, f->p4_s_list, "p4");
	gp_list_dump(m, f->gp_list);
    }
}

void scale_list_dump(M_P m, SC_P sc, char *p)
{
    for ( ; sc != NULL; sc = sc->next ) {
	fprintf(m->log,
	  "    Scale:  %s value %-4.8g frequency %-4.8g\n",
	  p, sc->value, sc->freq);
    }
}

void gp_list_dump(M_P m, GP_P gp)
{
    for ( ; gp != NULL; gp = gp->next ) {
	fprintf(m->log,
	  "Graph point:  frequency %-5.8g  gain %-4.8g\n",
	  gp->freq, gp->gain);
    }
}

void formant_list_dump(M_P m, FO_P f)
{
    for ( ; f != NULL; f = f->next ) {
	fprintf(m->log,
	  "  Formant:  type %1.1d p1 %-4.8g p2 %-4.8g p3 %-4.8g p4 %-4.8g\n",
	  f->type, f->p1, f->p2, f->p3, f->p4);
	gp_list_dump(m, f->gp_list);
    }
}

void wave_split_list_dump(M_P m, WP_P wp)
{
    for ( ; wp != NULL; wp = wp->next ) {
	fprintf(m->log, "  Wave Split  notes %d\n", wp->notes);
	wave_split_note_list_dump(m, wp->note_list);
	wave_split_condition_list_dump(m, wp->wpc_list);
    }
}

void wave_split_note_list_dump(M_P m, N_P n)
{
    for ( ; n != NULL; n = n->next ) {
	fprintf(m->log,
	  "    center %3.3d low %-3.3d high %-3.3d \n",
	  n->center, n->low, n->high);
    }
}

void wave_split_condition_list_dump(M_P m, WPC_P wpc)
{
    for ( ; wpc != NULL; wpc = wpc->next ) {
	fprintf(m->log,
	  "    relation %1.1d harmonic %-4.4d frequency %-4.8g divisor %-2.2d "
	  "remainder %-2.2d\n", wpc->rel, wpc->harmonic, wpc->freq,
	  wpc->divisor, wpc->remainder);
    }
}

/***********************************************************************

wave_list_dump

***********************************************************************/

/* forward references */
void sample_list_dump(M_P, S_P);
void harmonic_list_dump(M_P, H_P);

void wave_list_dump(M_P m)
{
    W_P ptr;

    fprintf(m->log, "\n\nwave_list_dump\n\n");
    for ( ptr = m->w_list; ptr != NULL; ptr = ptr->next ) {
	fprintf(m->log, 
	  "Wave Number %4.4d  Category %2.2s  Name %-8.8s  Layers %3.3d\n",
	  ptr->no, ptr->cat, ptr->name, ptr->layers);
	sample_list_dump(m, ptr->s_list);
    }
}

void sample_list_dump(M_P m, S_P s)
{
    for ( ; s != NULL; s = s->next ) {
	fprintf(m->log, "  Sample No %4.4d Cat %2.2s Name %-8.8s Freq %4.8g " 
	  "Ctr %3.3d Low %3.3d High %3.3d\n   Max Abs. Value %4.8g\n",
	   s->no, s->cat, s->name, s->freq, s->center, s->low, s->high,
	   s->max_absv);
	harmonic_list_dump(m, s->h_list);
    }
}

void harmonic_list_dump(M_P m, H_P h)
{
    for ( ; h != NULL; h = h->next ) {
     fprintf(m->log, "    Harmonic Number %4.4d  Level %10.6g  Phase %10.6g\n",
	  h->no, h->level, h->phase);
    }
}

/***********************************************************************

logws - log generated waves/sample/harmonics in input format
	adjusted for max 0 level for each sample

***********************************************************************/

/* forward references */

void logws(M_P m, double v_max_absv)
{
    WS_P ws_ptr; WP_P wp_ptr; W_P w_ptr; S_P s_ptr; H_P h_ptr;
    int gencnt, wavecnt;
    double v_norm, norm;

    /* convert v_max_absv from amplitude to dba */
    v_norm = 20 * log10(v_max_absv);

    /* wave splits correspond to generated waves, in order, so count 'em */
    gencnt = 0;

    /* for each wave set */
    for ( ws_ptr=m->ws_list; ws_ptr!=NULL; ws_ptr=ws_ptr->next ) {

	/* for each wave split */
	for ( wp_ptr=ws_ptr->wp_list; wp_ptr!=NULL; wp_ptr=wp_ptr->next ) {
	    gencnt++;
	}
    }

    /* log gencnt waves */
    fprintf(m->log, "\n\n*** generated waves for voice %s ***\n\n", m->v_name);
    wavecnt = 0;

    /* for each wave */
    for ( w_ptr=m->w_list; w_ptr!=NULL && wavecnt<gencnt; w_ptr=w_ptr->next ) {
	wavecnt++;
	fprintf(m->log,"wave\n");

	/* for each sample */
	for ( s_ptr=w_ptr->s_list; s_ptr!=NULL; s_ptr=s_ptr->next ) {
	    fprintf(m->log, "sample %d %d %d\n",
	      s_ptr->center, s_ptr->low, s_ptr->high);

	    /* get correct normalization factor */
	    if ( m->v_norm == 2 ) norm = v_norm;
	    else		  norm = 20 * log10(s_ptr->max_absv);
	    fprintf(m->log, "# Normalized by %g .\n\n", norm);

	    /* for each harmonic, output with adjusted level */
	    for ( h_ptr=s_ptr->h_list; h_ptr!=NULL; h_ptr=h_ptr->next ) {
		fprintf(m->log, "harmonic %d %g %g\n",
		  h_ptr->no, h_ptr->level - norm, h_ptr->phase);
	    }
	}
    }
    fprintf(m->log, "\n\n*** end generated waves for voice %s ***\n\n",
      m->v_name);
}

/***********************************************************************

plotwave - generate command file for gnuplot

***********************************************************************/

void plotwave(M_P m, W_P w_ptr, double v_max_absv)
{
    double v_norm, norm, freq;
    FILE *plotfile;
    S_P  s_ptr;
    H_P  h_ptr;
    int  scnt, xhi, yhi, i;
    char fname[14];

    /* convert v_max_absv from amplitude to dba */
    v_norm = 20 * log10(v_max_absv);

    /* count number of samples */
    scnt = 0;
    for ( s_ptr = w_ptr->s_list; s_ptr != NULL; s_ptr = s_ptr->next) scnt++;

    /* calculate some command parameters */
    if ( m->o_plotfreq ) xhi = 25000;
    else		 xhi = 11;
    if ( scnt == 1 )    yhi = 2;
    else		yhi = scnt;

    /* generate file name from wave name and open */
    strcpy(fname, w_ptr->name);
    strcat(fname, ".plt");
    plotfile = fopen(fname, "w");
    if ( plotfile == NULL ) errex(m, 0, "Cannot open plot file.");

    /* write initial commands */
    if ( m->o_plotfreq ) fprintf(plotfile, "set xlabel 'Frequency'\n");
    else 		     fprintf(plotfile, "set xlabel 'Octave'\n");
    fprintf(plotfile, "set ylabel 'Sample'\n");
    fprintf(plotfile, "set zlabel 'Level'\n");
    fprintf(plotfile, "set ticslevel 0\n");
    fprintf(plotfile, "set border 4095\n");
    fprintf(plotfile, "set grid xtics ytics ztics\n");
    fprintf(plotfile, "set view 50, 10, 1, 1\n");
    if ( m->o_plotfreq ) fprintf(plotfile,
      "set title 'Samples / Frequencies / Levels - %s'\n", w_ptr->name);
    else fprintf(plotfile,
      "set title 'Samples / Octaves / Levels - %s'\n", w_ptr->name);
    fprintf(plotfile, "splot[0:%d][1:%d][-100:0]\\\n", xhi, yhi);

    /* generate command for each sample */
    for ( i = 1; i < scnt; i++ ) 
	fprintf(plotfile, "'-' notitle with linespoints,\\\n");
    fprintf(plotfile, "'-' notitle with linespoints\n");
    
    /* for each sample */
    scnt = 0;
    for ( s_ptr = w_ptr->s_list; s_ptr != NULL; s_ptr = s_ptr->next) {
	scnt++;

	/* get correct normalization factor */
	if ( m->v_norm == 2 ) norm = v_norm;
	else		      norm = 20 * log10(s_ptr->max_absv);

	/* for each harmonic */
	for ( h_ptr = s_ptr->h_list; h_ptr != NULL; h_ptr = h_ptr->next ) {
	    if ( m->o_plotfreq )
	      freq = (s_ptr->freq)*(h_ptr->no);
	    else
	      freq = (log((s_ptr->freq)*(h_ptr->no))/log(2))-4;
	    fprintf(plotfile, "%g %d %g\n", freq, scnt, h_ptr->level - norm);
	}
	fprintf(plotfile, "e\n");
    }
    if ( m->o_plotpause )
      fprintf(plotfile, "pause -1 \"hit return to exit\"\n");

    fclose(plotfile);
}

/* end of misc.c */

