/**************************************************************

sscsss.c - simple single cycle sample set synthesizer
 (Say that 20 times in 1 minute!)
 generate a Yamaha EX series s1m (sample and wave) file
 per harmonic spectrum descriptions

This file is a part of sscsss -- Simple Single Cycle Sample Set Synthesizer.

Copyright (C) 2002 David Baker

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

**************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "sscsss.h"
#include "s1m.h"
#include "proto.h"

/**************************************************************

main

**************************************************************/

int main(int argc, char *argv[])
{
    double *sample, max_absv, v_max_absv;
    short  *isample;
    int    nsamples, i;
    char   filename[200], outfilename[200]; 
    M_S m;
    W_P w_ptr;
    S_P s_ptr;
    const char *initpath = "init.s1m";

    /* check command line args */
    if ( argc < 2 ) { printf("Usage:  sscsss file-name <init.s1m path>\n"); exit(0); }
    if ( argc > 2 ) initpath = argv[2];


    m.log = NULL;	/* errors to stdout until log file opened */
    m.w_back = NULL;	/* keep wave out of initial error messages */

    /* open input file */
    m.infile = fopen(argv[1], "r");
    if ( m.infile == NULL ) errex(&m, 0, "Cannot open input file.");

    /* initialize options */
    m.o_lowfreq	  = 1.0;
    m.o_mindba    = -60.0;
    m.o_debug	  = 0;
    m.o_logws	  = 0;
    m.o_mkplot	  = 0;
    m.o_plotfreq  = 0;
    m.o_plotpause = 0;

    /* get options and output file name */
    m.linecnt	= 0;
    m.state	= IS_INIT;
    get_opts_fname(&m, filename);
    
    /* construct output file name, log, and open */
    strcpy(outfilename, filename);
    strcat(outfilename, ".s1m");
    fprintf(m.log, "Output File:  %s\n", outfilename);
    m.outfile = fopen(outfilename, "wb");
    if ( m.outfile == NULL ) errex(&m, 1, "Cannot open outputfile (s1m).");

    /* allocate memory for sample point conversion */
    nsamples = (int) ceil(S1M_SR / m.o_lowfreq) + 2;
    isample = (short *) malloc(sizeof(short) * nsamples);
    if ( isample == NULL ) errex(&m, 0, "Unable to allocate sample memory.");

    /* main loop, read input file and process by voice */

    /* initialize for main loop */
    m.f_size	= 0;
    m.f_waves	= 0;
    m.f_samples	= 0;
    m.w_list	= NULL; 
    m.ws_list	= NULL; 
    m.d_malloc	= 0;
    m.d_free	= 0;
    init_s1m_header(&m, initpath);
    write_s1m_header(&m);

    /* for each voice in input file */
    while ( m.state != IS_EOF ) {
	v_max_absv = 0.0;
	get_voice(&m);

	/* pass 1 calculate sample data and max absolute value */

	/* for each wave in voice */
	for ( w_ptr = m.w_list; w_ptr != NULL; w_ptr = w_ptr->next ) {

	    /* for each sample in wave */
	    for ( s_ptr = w_ptr->s_list; s_ptr != NULL; s_ptr = s_ptr->next ) {
		sample = (double *) malloc(sizeof(double) * s_ptr->smpl_len);
		(m.d_malloc)++;
		if ( sample == NULL )
		  errex(&m, 0, "Unable to allocate sample memory.");
		gen_sample(s_ptr, sample, &max_absv);
		if ( max_absv > v_max_absv ) v_max_absv = max_absv;
		s_ptr->max_absv = max_absv;
		s_ptr->data = sample;
	    }
	}

	/* pass 2 use data from pass 1 to do rest of the work */

	/* for each wave in voice */
	for ( w_ptr = m.w_list; w_ptr != NULL; w_ptr = w_ptr->next ) {
	    add_s1m_wave(&m, w_ptr);
	    if ( m.o_mkplot ) plotwave(&m, w_ptr, v_max_absv);

	    /* for each sample in wave */
	    for ( s_ptr = w_ptr->s_list; s_ptr != NULL; s_ptr = s_ptr->next ) {
		for ( i = 0; i < s_ptr->smpl_len; i++) { /* normalize data */
		    if ( m.v_norm == 2 ) s_ptr->data[i] /= v_max_absv;
		    else		 s_ptr->data[i] /= s_ptr->max_absv;
		}
		add_s1m_sample(&m, s_ptr);
		write_s1m_sample(&m, s_ptr, s_ptr->data, isample);
		add_s1m_layer(&m, s_ptr);
	    }
	}

	/* print debug info, if requested */
	if ( m.o_debug ) {
	    dump_m(&m);
	    wave_set_list_dump(&m);
	    wave_list_dump(&m);
	}
	/* log generated waves, if requested */
	if ( m.o_logws ) logws(&m, v_max_absv);

	/* free memory dynamically allocated for voice data */
	free_lists(&m);
    }

    /* finish up the s1m file */
    write_s1m_end(&m);
    rewind(m.outfile);
    write_s1m_header(&m);
    
    /* print malloc and free call counts, if requested */
    if ( m.o_debug ) {
	fprintf(m.log, "Calls to malloc:  %-9.9d .\n", m.d_malloc);
	fprintf(m.log, "Calls to free  :  %-9.9d .\n", m.d_free  );
    }
    /* log some statistics */
    fprintf(m.log, "Waves:             %4d .\n", m.f_waves);
    fprintf(m.log, "Samples:           %4d .\n", m.f_samples);
    fprintf(m.log, "Sample Bytes:  %8d .\n", m.f_size);
    
    /* close the files and exit */
    fclose(m.outfile);
    fclose(m.infile);
    fclose(m.log);
    exit(0);
}

/* end of sscsss.c */

