/************************************************************************

waveset.c - generate waves

This file is a part of sscsss -- Simple Single Cycle Sample Set Synthesizer.

Copyright (C) 2002 David Baker

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

************************************************************************/

#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "sscsss.h"
#include "s1m.h"
#include "proto.h"

/* forward references */
void   allocate_wave(M_P);
void   make_note_list(M_P, WS_P, WP_P);
void   allocate_sample(M_P, N_P);
int    not_too_high(M_P, int);
char   chk_wp_conds(M_P, int, WPC_P);
void   apply_offsets(M_P, WS_P, int, double *, double *);
void   apply_filters(M_P, WS_P, int, double *);
void   apply_formants(M_P, WS_P, int, double *);
void   allocate_harmonic(M_P, int, double, double);
double scale(M_P, SC_P, double);
double f_cos(double, double, double, double);
double f_tri(double, double, double, double);
double f_hpf(double, double, double);
double f_lpf(double, double, double);
double f_hs(M_P, double, double, double, double);
double f_ls(M_P, double, double, double, double);
double f_band(M_P, double, double, double, double, double);
double f_graph(M_P, double, GP_P, int);

void gen_waves(M_P m)
{
    double level, phase;
    int    h_no;
    WS_P   ws_ptr; WP_P wp_ptr; N_P nl_ptr;

    /* for each wave-set */
    for ( ws_ptr = m->ws_list; ws_ptr != NULL; ws_ptr = ws_ptr->next ) {

	/* for each wave-split */
	for ( wp_ptr=ws_ptr->wp_list; wp_ptr!=NULL; wp_ptr=wp_ptr->next ) {
	    allocate_wave(m);
	    make_note_list(m, ws_ptr, wp_ptr);

	    /* for each sample */
	    for ( nl_ptr=wp_ptr->note_list;nl_ptr!=NULL;nl_ptr=nl_ptr->next ) {
		allocate_sample(m, nl_ptr);

		/* for each harmonic below half sampling rate */
		for ( h_no = 1; not_too_high(m, h_no); h_no++ ) {

		    /* if wave-split conditions for including are met */
		    if ( chk_wp_conds(m, h_no, wp_ptr->wpc_list) ) {

			/* generate harmonic */
			level = 0.0;
			phase = 0.0;
			apply_offsets(m, ws_ptr, h_no, &level, &phase);
			apply_filters(m, ws_ptr, h_no, &level);
			apply_formants(m, ws_ptr, h_no, &level);
			if ( level >= m->o_mindba )
			  allocate_harmonic(m, h_no, level, phase);
		    }
		}
		if ( m->h_back==NULL ) errex(m,0,"Sample with no harmonics.");
		m->h_back = NULL;
	    }
	    m->s_back = NULL;
	    m->w_back->layers = m->w_samples;
	    m->w_samples = 0;
	    m->n_back=NULL;
	}
    }
}

/************************************************************************

allocate_wave

************************************************************************/

void allocate_wave(M_P m)
{
    char *wavnos[] = { "1", "2", "3", "4" };
    W_P ptr;

    /* check number of waves for voice and file */
    if ( m->v_waves+1 > S1M_WPV ) errex(m, 0, "Too many waves in voice.");
    if ( m->f_waves+1 > S1M_WPF ) errex(m, 0, "Too many waves in file.");

    /* input ok */

    /* allocate new wave structure and put in list */
    ptr = (W_P) malloc (sizeof(W_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 0, "Unable to allocate wave struct.");
    if ( m->w_back == NULL ) m->w_list = ptr;
    else m->w_back->next = ptr;
    m->w_back = ptr;

    /* increment wave counts for voice and file */
    m->v_waves++;
    m->f_waves++;

    /* initialize data for new wave structure */
    ptr->no = m->f_waves;                       /* wave number   */
    strcpy(ptr->cat, m->v_cat);                 /* wave category */
    strncpy(ptr->name, m->v_name, 5);           /* wave name     */
    ptr->name[5] = '\0';
    strcat(ptr->name, wavnos[m->v_waves-1]);
    ptr->next   = NULL;				/* next wave     */
    ptr->s_list = NULL;				/* sample list   */

    /* debug message */
    if ( m->o_debug ) fprintf(m->log, "Allocated wave %s.\n", ptr->name);
}

/************************************************************************

make_note_list

************************************************************************/

void make_note_list(M_P m, WS_P ws_ptr, WP_P wp_ptr)
{
    int center;
    N_P nl_ptr;

    /* if a list exists already, return */
    if ( wp_ptr->note_list != NULL ) return;

    /* else generate note list using voice low high, split number */
    else if ( wp_ptr->notes != 0 ) {
	for ( center=m->v_low; center <= m->v_high; center+=wp_ptr->notes ) {

	    /* allocate new note structure and put in list */
	    nl_ptr = (N_P) malloc (sizeof(N_S));
	    (m->d_malloc)++;
	    if ( nl_ptr == NULL ) errex(m,0,"Unable to allocate note struct.");
	    if ( m->n_back == NULL ) wp_ptr->note_list = nl_ptr;
	    else m->n_back->next = nl_ptr;
	    m->n_back = nl_ptr;

	    /* initialize data */
	    nl_ptr->center = center;
	    nl_ptr->low    = center;
	    nl_ptr->high   = center + wp_ptr->notes - 1;
	    if ( nl_ptr->high > m->v_high ) nl_ptr->high = m->v_high;
	    nl_ptr->next   = NULL;
	}
    }
    /* else don't know what to do */
    else errex(m, 0, "Notes = 'list' and no note list, nothing to generate.");
}

/************************************************************************

allocate_sample

************************************************************************/

void allocate_sample(M_P m, N_P nl_ptr)
{
    char hconv[] = "0123456789abcdef";
    int snlen;
    double freq;
    S_P ptr;
    double per;

    /* calculate freq */
    freq = 440.0 * pow(pow(2.0, 1.0 / 12.0), nl_ptr->center - 69);

    /* check number of samples per wave and file */
    if ( m->w_samples+1 > S1M_SPW ) errex(m, 0, "Too many samples in wave.");
    if ( m->f_samples+1 > S1M_SPF ) errex(m, 0, "Too many samples in file.");

    /* allocate new sample structure and put in list */
    ptr = (S_P) malloc (sizeof(S_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 0, "Unable to allocate sample struct.");
    if ( m->s_back == NULL ) m->w_back->s_list = ptr;
    else (m->s_back)->next = ptr;
    m->s_back = ptr;

    /* update m */
    m->w_samples += 1;
    m->f_samples += 1;
    m->h_no = 0;

    /* set data in structure */
    ptr->no = m->f_samples;             /* sample number */

    strcpy(ptr->cat, m->w_back->cat);   /* copy sample category from wave */

    strcpy(ptr->name, m->w_back->name); /* generate sample name */
    snlen = strlen(ptr->name);
    ptr->name[snlen    ]=hconv[(nl_ptr->center>>4)&0x0f]; /* append hex of */
    ptr->name[snlen + 1]=hconv[ nl_ptr->center    & 0x0f]; /* center */
    ptr->name[snlen + 2]='\0';

    ptr->freq   = freq;
    ptr->center = nl_ptr->center;
    ptr->low    = nl_ptr->low;
    ptr->high   = nl_ptr->high;
    if ( nl_ptr->high > m->v_high ) ptr->high = m->v_high;

    /* additional calculations for sample generation */
    per = S1M_SR / freq;
    ptr->loop_len = (int) floor(per) + 1;
    ptr->loop_fract = 64 - (int) round((per - floor(per)) * 64);
    if ( ptr->loop_fract == 64 ) {
	ptr->loop_fract = 0;
	ptr->loop_len -= 1;
    }
    ptr->smpl_len = ptr->loop_len + 2;

    /* initialize pointers to next (if any) structure */
    ptr->next   = NULL;
    ptr->h_list = NULL;

    /* debug message */
    if ( m->o_debug ) fprintf(m->log, "Allocated sample %s.\n", ptr->name);
}

/************************************************************************

not_too_high - aliasing check

************************************************************************/

int not_too_high(M_P m, int hno)
{
    double freq;

    freq = m->s_back->freq * hno *
      pow(pow(2.0, 1.0/12.0),(m->s_back->high - m->s_back->center)+m->v_bend);
    if ( freq > S1M_SR / 2.0 ) return 0;
    else return 1;
}

/************************************************************************

chk_wp_conds

************************************************************************/

#define STACKSIZE 10

/* forward references */
void push(M_P, char *, int *, char);
char pop(M_P, char *, int *);
char stktop(M_P, char *, int);
char chk_cond(M_P, int, WPC_P);

char chk_wp_conds(M_P m, int h_no, WPC_P wpc_ptr)
{
    char include, logicalop, t1, t2;
    char stack[STACKSIZE]; /* only logical operands on stack */
    WPC_P wpc_ptr1;

    /* check for logical operators */
    logicalop = 0;
    for ( wpc_ptr1 = wpc_ptr; wpc_ptr1 != NULL; wpc_ptr1 = wpc_ptr1->next ) {
	if ( wpc_ptr1->rel == SP_C_NOT ||
	     wpc_ptr1->rel == SP_C_AND ||
	     wpc_ptr1->rel == SP_C_OR ) { logicalop = 1; break;}
    }

    /* process as rpn list */
    if ( logicalop ) {
	int top = -1; /* empty */
	

	for ( wpc_ptr1=wpc_ptr; wpc_ptr1!=NULL; wpc_ptr1=wpc_ptr1->next ) {
	    switch ( wpc_ptr1->rel ) {
		case SP_C_NOT: {
		    t1 = pop(m, stack, &top);
		    push(m, stack, &top, ! t1);
		    break;
		}
		case SP_C_AND: {
		    t1 = pop(m, stack, &top);
		    t2 = pop(m, stack, &top);
		    push(m, stack, &top, t1 && t2);
		    break;
		}
		case SP_C_OR: {
		    t1 = pop(m, stack, &top);
		    t2 = pop(m, stack, &top);
		    push(m, stack, &top, t1 || t2 );
		    break;
		}
		default: {
		    t1 = chk_cond(m, h_no, wpc_ptr1);
		    push(m, stack, &top, t1);
		    break;
		}
	    }
	}
	if ( top != 0 ) errex(m, 0,
	  "Condition list finished with more than 1 operand on stack.");
	include = stktop(m, stack, top);
    }

    /* process as and list */
    else {
	/* for each condition */
	include = 1;
	for ( wpc_ptr1=wpc_ptr; wpc_ptr1!=NULL; wpc_ptr1=wpc_ptr1->next ) {

	    /* if not condition */
	    if ( ! chk_cond(m, h_no, wpc_ptr1) ) { include = 0; break; }
	}
    }

    return include;
}

/************************************************************************

stack functions

************************************************************************/

void push(M_P m, char *stack, int *top, char val)
{
    if ( *top + 2 > STACKSIZE ) errex(m, 0, "Stack overflow.");
    stack[++(*top)] = val;
}

char pop(M_P m, char *stack, int *top)
{
    if ( *top == -1 ) errex(m, 0, "Stack underflow.");
    return stack[(*top)--];
}

char stktop(M_P m, char *stack, int top)
{
    if ( top == -1 ) errex(m, 0, "Empty stack.");
    return stack[top];
}

/************************************************************************

chk_cond

************************************************************************/

char chk_cond(M_P m, int h_no, WPC_P wpc_ptr)
{
    int ok;

    ok = 0;

    if      ( wpc_ptr->rel == SP_C_EQ ) {	/* =  */
	if ( wpc_ptr->freq == -1.0 ) {if ( h_no == wpc_ptr->harmonic ) ok = 1;}
	else {if ( h_no * m->s_back->freq == wpc_ptr->freq ) ok = 1;}
    }
    else if ( wpc_ptr->rel == SP_C_LT ) {	/* <  */
	if ( wpc_ptr->freq == -1.0 ) {if ( h_no < wpc_ptr->harmonic ) ok = 1;}
	else {if ( h_no * m->s_back->freq < wpc_ptr->freq ) ok = 1;}
    }
    else if ( wpc_ptr->rel == SP_C_LE ) {	/* <= */
	if ( wpc_ptr->freq == -1.0 ) {if ( h_no <= wpc_ptr->harmonic ) ok = 1;}
	else {if ( h_no * m->s_back->freq <= wpc_ptr->freq ) ok = 1;}
    }
    else if ( wpc_ptr->rel == SP_C_GT ) {	/* >  */
	if ( wpc_ptr->freq == -1.0 ) {if ( h_no > wpc_ptr->harmonic ) ok = 1;}
	else {if ( h_no * m->s_back->freq > wpc_ptr->freq ) ok = 1;}
    }
    else if ( wpc_ptr->rel == SP_C_GE ) {	/* >= */
	if ( wpc_ptr->freq == -1.0 ) {if ( h_no >= wpc_ptr->harmonic ) ok = 1;}
	else {if ( h_no * m->s_back->freq >= wpc_ptr->freq ) ok = 1;}
    }
    else if ( wpc_ptr->rel == SP_C_MOD ) {	/* %  */
	if ( (h_no % wpc_ptr->divisor) == wpc_ptr->remainder ) ok = 1;
    }

    return ok;
}

/************************************************************************

apply_offsets

************************************************************************/

void apply_offsets(M_P m, WS_P ws_ptr, int hno, double *level, double *phase)
{
    O_P o_ptr;

    for ( o_ptr = ws_ptr->o_list; o_ptr != NULL; o_ptr = o_ptr->next ) {
	    
	if ( o_ptr->no == hno ) {

	    if ( o_ptr->level_s_list   != NULL ) {/* if level  scaled */
		*level += scale(m, o_ptr->level_s_list, o_ptr->level);
	    }
	    else *level += o_ptr->level;
	    
	    if ( o_ptr->phase_s_list != NULL ) {/* if phase  scaled */
		*phase += scale(m, o_ptr->phase_s_list, o_ptr->phase);
	    }
	    else *phase += o_ptr->phase;
	}
    }
}

/************************************************************************

apply_filters

************************************************************************/

void apply_filters(M_P m, WS_P ws_ptr, int hno, double *level)
{
    FI_P f_ptr;
    double h_freq, p1, adjp1, p2, p3, p4;

    if ( ws_ptr->fi_list == NULL ) return;	/* no filters to apply */

    /* calculate log base 2 of harmonic frequency (convert to octaves) */
    h_freq = log(hno * m->s_back->freq) / log(2);

    /* for each filter */
    for ( f_ptr = ws_ptr->fi_list; f_ptr != NULL; f_ptr = f_ptr->next ) {
	    
	if ( f_ptr->p1_s_list != NULL ) {/* if p1 scaled */
	    p1 = scale(m, f_ptr->p1_s_list, f_ptr->p1);
	}
	else p1 = f_ptr->p1;
	adjp1 = p1 + log(m->s_back->freq) / log(2); /* adjust to fundamental */
	
	if ( f_ptr->p2_s_list   != NULL ) {/* if p2 scaled */
	    p2 = scale(m, f_ptr->p2_s_list, f_ptr->p2);
	}
	else p2 = f_ptr->p2;
	
	if ( f_ptr->p3_s_list  != NULL ) {/* if p3 scaled */
	    p3 = scale(m, f_ptr->p3_s_list, f_ptr->p3);
	}
	else p3 = f_ptr->p3;

	if ( f_ptr->p4_s_list  != NULL ) {/* if p4 scaled */
	    p4 = scale(m, f_ptr->p4_s_list, f_ptr->p4);
	}
	else p4 = f_ptr->p4;

	switch (f_ptr->type) {
	    case F_COS:
		*level += f_cos(h_freq, adjp1, p2, p3);
		break;
	    case F_TRI:
		*level += f_tri(h_freq, adjp1, p2, p3);
		break;
	    case F_HPF:
		*level += f_hpf(h_freq, adjp1, p2);
		break;
	    case F_LPF:
		*level += f_lpf(h_freq, adjp1, p2);
		break;
	    case F_LPF_F:
		*level += f_lpf(hno, p1, p2);
		break;
	    case F_HS:
		*level += f_hs(m, h_freq, adjp1, p2, p3);
		break;
	    case F_LS:
		*level += f_ls(m, h_freq, adjp1, p2, p3);
		break;
	    case F_BAND:
		p2 += log(m->s_back->freq) / log(2); /* adjust to fund. */
		*level += f_band(m, h_freq, adjp1, p2, p3, p4);
		break;
	    case F_GRAPH:
		*level += f_graph(m, h_freq, f_ptr->gp_list, 1);
		break;
	}
    }
}

/************************************************************************

apply_formants

************************************************************************/

void apply_formants(M_P m, WS_P ws_ptr, int hno, double *level)
{
    FO_P f_ptr;
    double h_freq, c_freq, freq2;

    if ( ws_ptr->fo_list == NULL ) return;  /* no formant's to apply */
    f_ptr = ws_ptr->fo_list;

    /* calculate log base 2 of harmonic frequency (convert to octaves) */
    h_freq = log(hno * m->s_back->freq) / log(2);

    /* for each formant */
    for ( ; f_ptr != NULL; f_ptr = f_ptr->next ) {

	/* calculate log base 2 of formant center frequency */
	c_freq  = log(f_ptr->p1) / log(2);

	switch (f_ptr->type) {
	    case F_COS:
		*level += f_cos(h_freq, c_freq, f_ptr->p2, f_ptr->p3);
		break;
	    case F_TRI:
		*level += f_tri(h_freq, c_freq, f_ptr->p2, f_ptr->p3);
		break;
	    case F_HPF:
		*level += f_hpf(h_freq, c_freq, f_ptr->p2);
		break;
	    case F_LPF:
		*level += f_lpf(h_freq, c_freq, f_ptr->p2);
		break;
	    case F_HS:
		*level += f_hs(m, h_freq, c_freq, f_ptr->p2, f_ptr->p3);
		break;
	    case F_LS:
		*level += f_ls(m, h_freq, c_freq, f_ptr->p2, f_ptr->p3);
		break;
	    case F_BAND:
		freq2 = log(f_ptr->p2) / log(2);
		*level += f_band(m, h_freq,c_freq,freq2,f_ptr->p3,f_ptr->p4);
		break;
	    case F_GRAPH:
		*level += f_graph(m, h_freq, f_ptr->gp_list, 0);
		break;
	}
    }
}

/************************************************************************

allocate_harmonic

************************************************************************/

void allocate_harmonic(M_P m, int hno, double level, double phase)
{
    H_P ptr;

    /* allocate new harmonic structure and put in list */
    ptr = (H_P) malloc (sizeof(H_S));
    (m->d_malloc)++;
    if ( ptr == NULL ) errex(m, 0, "Unable to allocate harmonic struct.");
    if ( m->h_back == NULL ) m->s_back->h_list = ptr;
    else (m->h_back)->next = ptr;
    m->h_back = ptr;

    /* set data in structure */
    m->h_no    = hno;
    ptr->no    = m->h_no;
    ptr->level = level;
    ptr->phase = phase;
    ptr->next  = NULL;
}

/************************************************************************

scale - linear interpolation on log base 2 of frequency

************************************************************************/

double scale(M_P m, SC_P sc_ptr, double initv)
{
    double x1, y1;	/* freq value for init or lower scale point */
    double x2, y2;	/* freq value for upper scale point */
    double x, y;	/* freq (known) value (unknown) for current sample */

    if ( m->s_back->low == m->v_low ) return initv;

    x  = m->s_back->freq;

    x1 = 440.0 * pow(pow(2.0, 1.0 / 12.0), m->v_low - 69);
    y1 = initv;
    for ( ; sc_ptr != NULL; sc_ptr = sc_ptr->next ) {
	x2 = sc_ptr->freq;
	y2 = sc_ptr->value;
	if ( x2 > x ) break;
	x1 = x2;
	y1 = y2;
    }

    if ( x1 == x2 ) return y2;

    x1 = log(x1) / log(2);
    x2 = log(x2) / log(2);
    x  = log(x ) / log(2);
    y = y1 + ((x-x1) / (x2-x1)) * (y2-y1);

    return y;
}

/************************************************************************

f_cos - sinusoid with offset center, 0 at width, amplitude of gain

************************************************************************/

#define PI 3.14159265358979323846

double f_cos(double x, double center, double gain, double width)
{
    if ( x < center - .5 * width ) return 0.0;
    if ( x > center + .5 * width ) return 0.0;
    return gain*.5*( cos((1.0/width)*2.0*PI*x - (center/width)*2.0*PI) +1);
}

/************************************************************************

f_tri - triangle with offset center, 0 at width, amplitude of gain

************************************************************************/

double f_tri(double x, double center, double gain, double width)
{
    if ( x < center - .5 * width ) return 0.0;
    if ( x > center + .5 * width ) return 0.0;
    if ( x == center ) return gain;
    if ( x < center )
	return (((x-(center-.5*width))/(.5*width))*(gain-0.0))+0.0; 
	return (((x-(center))         /(.5*width))*(0.0-gain))+gain;
}

/************************************************************************

f_hpf - high pass filter

************************************************************************/

double f_hpf(double x, double cutoff, double slope)
{
    if ( x >= cutoff ) return 0.0;
    return (cutoff - x) * slope;
}

/************************************************************************

f_lpf - low pass filter

************************************************************************/

double f_lpf(double x, double cutoff, double slope)
{
    if ( x <= cutoff ) return 0.0;
    return (x - cutoff) * slope;
}

/************************************************************************

f_hs - high shelf filter

************************************************************************/

double f_hs(M_P m, double x, double freq, double gain, double slope)
{
    double tmp;

    /* check for scaling violating gain and slope having same sign */
    if ( ( gain < 0 && slope > 0 ) || ( gain > 0 && slope < 0 ) )
      errex(m, 0, "High shelf gain and slope have different signs, scaling?");

    if ( x <= freq ) return 0.0;
    tmp = (x - freq) * slope;
    if ( fabs(tmp) > fabs(gain) ) return gain;
    return tmp;
}

/************************************************************************

f_ls - low shelf filter

************************************************************************/

double f_ls(M_P m, double x, double freq, double gain, double slope)
{
    double tmp;

    /* check for scaling violating gain and slope having same sign */
    if ( ( gain < 0 && slope > 0 ) || ( gain > 0 && slope < 0 ) )
      errex(m, 0, "Low shelf gain and slope have different signs, scaling?");

    if ( x >= freq ) return 0.0;
    tmp = (freq - x) * slope;
    if ( fabs(tmp) > fabs(gain) ) return gain;
    return tmp;
}

/************************************************************************

f_band - band pass/elimination filter

************************************************************************/

double f_band(M_P m, double x, double f1, double f2, double g, double s)
{
    double tmp;

    /* check for scaling setting freq1 > freq2 */
    if ( f1 > f2 ) errex(m, 0, "Freq1 > freq2 in band filter, scaling?");
    
    if ( x >= f1 && x <= f2 ) return g;

    if ( x < f1 ) {
	if ( g > 0.0 ) {
	    tmp = g - (f1 - x) * s;
	    if ( tmp < 0.0 ) tmp = 0.0;
	}
	else {
	    tmp = g + (f1 - x) * s;
	    if ( tmp > 0.0 ) tmp = 0.0;
	}
    }
    else {
	if ( g > 0.0 ) {
	    tmp = g - (x - f2) * s;
	    if ( tmp < 0.0 ) tmp = 0.0;
	}
	else {
	    tmp = g + (x - f2) * s;
	    if ( tmp > 0.0 ) tmp = 0.0;
	}
    }
    return tmp;
}

/************************************************************************

f_graph - graphic (multiple point with interpolation) filter

************************************************************************/

double f_graph(M_P m, double h_freq, GP_P g_ptr, int filter)
{
    double adj, x1, y1, x2, y2;

    if ( g_ptr == NULL ) return 0.0;	/* nothing to do */

    if ( filter ) adj = log(m->s_back->freq) / log(2);
    else	  adj = 0.0;
    
    x2 = -1.0; y2 = -1.0;
    for ( ; g_ptr != NULL; g_ptr = g_ptr->next ) {
	x1 = x2;
	y1 = y2;
	if ( filter ) x2 = adj + g_ptr->freq;
	else	      x2 = adj + log(g_ptr->freq) / log(2);
	y2 = g_ptr->gain;
	if ( x2 >= h_freq ) break;
    }

    if ( x1 == -1.0 || h_freq >= x2 ) return y2;
    
    return y1 + ((h_freq - x1) / (x2 - x1)) * (y2 - y1);
}

/* end of waveset.c */

